/* 
 * $Id: sysctl.c,v 1.17 2001/10/23 00:14:23 thockin Exp $
 *
 * Copyright (c) 2001, Sun Microsystems, Inc.
 * All Rights Reserved
 *
 * A melting pot for cobalt sysctls
 */

#include <linux/config.h>
#ifdef CONFIG_COBALT

#include <linux/string.h>
#include <linux/sysctl.h>
#include <linux/nvram.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/cobalt.h>
#include <linux/cobalt-nvram.h>
#include <linux/cobalt-led.h>
#include <linux/cobalt-systype.h>
#include <linux/cobalt-superio.h>

/* sysfault */
static int sysfault_status;
static int do_sysfault(ctl_table *table, int *name, int nlen,
	void *oldval, size_t *oldlenp, void *newval, size_t newlen, 
	void **context);
static int proc_dosysfault(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp);
static int set_sysfault(int val);
static int set_sysfault_led(int val);
static int set_sysfault_nvram(int val);
static void init_sysfault(void);

/* powermode */
static enum cobalt_powermode powermode_status;
static int do_powermode(ctl_table *table, int *name, int nlen,
	void *oldval, size_t *oldlenp, void *newval, size_t newlen, 
	void **context);
static int proc_dopowermode(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp);
static int set_powermode(int val);
static int powermode_fail(const int set, const int val);
static int powermode_apply(const int set, const int val);
static void init_powermode(void);

/* the table of all cobalt sysctl()s */
ctl_table cobalt_table[] = {
	{ 
		ctl_name: COBALT_SYSFAULT,
		procname: "sysfault",
		data: &sysfault_status, 
		maxlen: sizeof(sysfault_status),
		mode: 0640, 
		child: NULL, 
		proc_handler: proc_dosysfault,
		strategy: do_sysfault,
	},
	{
		ctl_name: COBALT_POWERMODE,
		procname: "powermode",
		data: &powermode_status, 
		maxlen: sizeof(powermode_status),
		mode: 0640, 
		child: NULL, 
		proc_handler: proc_dopowermode,
		strategy: do_powermode
	},
	{0}
};

/*
 * Cobalt Sysfault flag
 */
static int 
do_sysfault(ctl_table *table, int *name, int nlen,
	void *oldval, size_t *oldlenp, void *newval, size_t newlen, 
	void **context)
{
	/* handle a write trigger, else let the automagic stuff do it */
	if (newval && newlen) {
		int val;
		if (get_user(val, (int *)newval) < 0) {
			return -EFAULT;
		}
		return set_sysfault(val);
	}
	return 0;
}

static int 
proc_dosysfault(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	int err;
	int oldval = sysfault_status;

	/* first call proc_dointvec to parse the input/output */
	if ((err = proc_dointvec(table, write, filp, buffer, lenp)) < 0) {
		return err;
	}

	/* handle a write trigger */
	if (write) {
		err = set_sysfault(sysfault_status);
		if (err < 0) {
			/* proc_dointvec blows away the stored value */
			sysfault_status = oldval;
		}
	}
	return err;
}

static int 
set_sysfault(int val)
{
	if (val != 0 && val != 1) {
		return -EINVAL;
	}

	set_sysfault_led(val);
	set_sysfault_nvram(val);

	return 0;
}

static int
set_sysfault_led(int val)
{
#ifdef CONFIG_COBALT_LED
	if (val) {
		cobalt_led_set_bits_lazy(LED_SYSFAULT);
	} else {
		cobalt_led_clear_bits_lazy(LED_SYSFAULT);
	}
#endif
	return 0;
}

static int
set_sysfault_nvram(int val)
{
#ifdef CONFIG_NVRAM
	unsigned short nvram_flags;

	nvram_flags = (nvram_read_byte(COBT_CMOS_FLAG_BYTE_0) << 8) |
		nvram_read_byte(COBT_CMOS_FLAG_BYTE_1);

	if (val) {
		nvram_flags |= COBT_CMOS_SYSFAULT_FLAG;
	} else {
		nvram_flags &= ~COBT_CMOS_SYSFAULT_FLAG;
	}

	nvram_write_byte(nvram_flags & 0xff, COBT_CMOS_FLAG_BYTE_1);
	nvram_write_byte(nvram_flags >> 8, COBT_CMOS_FLAG_BYTE_0);
	nvram_set_checksum();
#endif
	return 0;
}

static void
init_sysfault(void)
{
#ifdef CONFIG_NVRAM
	unsigned short nvram_flags;

	/* grab the value of sysfault out of cmos */
	nvram_flags = (nvram_read_byte(COBT_CMOS_FLAG_BYTE_0) << 8) |
		nvram_read_byte(COBT_CMOS_FLAG_BYTE_1);
	
	if (nvram_flags & COBT_CMOS_SYSFAULT_FLAG) {
		sysfault_status = 1;
	} else {
		sysfault_status = 0;
	}
	set_sysfault(sysfault_status);
#endif
}

/*
 * Cobalt power-mode
 */
static int 
do_powermode(ctl_table *table, int *name, int nlen,
	void *oldval, size_t *oldlenp, void *newval, size_t newlen, 
	void **context)
{
	/* handle a write trigger, else let the automagic stuff do it */
	if (newval && newlen) {
		int val;
		if (get_user(val, (int *)newval) < 0) {
			return -EFAULT;
		}
		return set_powermode(val);
	}
	return 0;
}

static int 
proc_dopowermode(ctl_table *table, int write, struct file *filp,
	void *buffer, size_t *lenp)
{
	int err = 0;
	int oldval = powermode_status;
	char buf[16];
	int len = 0;

	if (filp->f_pos && !write) {
		*lenp = 0;
		return 0;
	}
	if (write) {
		/* read the string from userland */
		enum cobalt_powermode val;
		char *p;
		if (*lenp > sizeof(buf)-1) {
			len = sizeof(buf)-1;
		} else {
			len = *lenp;
		}
		if (copy_from_user(buf, buffer, len) < 0) {
			return -EFAULT;
		}
		buf[len] = '\0';
		/* if it is 0, 1, 2 handle it */
		val = simple_strtol(buf, &p, 0);
		if (p != buf) {
			/* got a number */
			powermode_status = val;
		} else {
			/* match it */
			if (buf[len-1] == '\n')
				buf[len-1] = '\0';
			if (!strcmp(buf, "on")) {
				powermode_status = COBALT_POWER_ON;
			} else if (!strcmp(buf, "off")) {
				powermode_status = COBALT_POWER_OFF;
			} else if (!strcmp(buf, "same")) {
				powermode_status = COBALT_POWER_SAME;
			} else {
				return -EINVAL;
			}
		}
		err = set_powermode(powermode_status);
		if (err < 0) {
			/* proc_dointvec blows away the stored value */
			powermode_status = oldval;
		}
	} else {
		/* build a buffer */
		switch (powermode_status) {
		case COBALT_POWER_ON:
			strcpy(buf, "on");
			break;
		case COBALT_POWER_OFF:
			strcpy(buf, "off");
			break;
		case COBALT_POWER_SAME:
			strcpy(buf, "same");
			break;
		default:
			strcpy(buf, "unknown");
		}
		strcat(buf, "\n");
		len = strlen(buf);
		/* copy it up */
		if (copy_to_user(buffer, buf, len) < 0) {
			return -EFAULT;
		}
		*lenp = len;
	}
	if (!err) {
		filp->f_pos += len;
	}

	return err;
}

#define cobt_get_powerfail()	powermode_fail(0, 0)
#define cobt_set_powerfail(a)	powermode_fail(1, (a))
#define cobt_get_powerapply()	powermode_apply(0, 0)
#define cobt_set_powerapply(a)	powermode_apply(1, (a))
static int
set_powermode(int val)
{
	int r0,r1;
	int fail, apply;
	
	fail = cobt_get_powerfail();
	apply = cobt_get_powerapply();

	switch (val) {
	case COBALT_POWER_ON:
		r0 = cobt_set_powerfail(1);
		r1 = cobt_set_powerapply(1);
		break;
	case COBALT_POWER_OFF:
		r0 = cobt_set_powerfail(0);
		r1 = cobt_set_powerapply(0);
		break;
	case COBALT_POWER_SAME:
		r0 = cobt_set_powerfail(1);
		r1 = cobt_set_powerapply(0);
		break;
	default:
		return -EINVAL;
	}

	if (r0 < 0 || r1 < 0) {
		/* one of them failed - roll it back */
		cobt_set_powerfail(fail);
		cobt_set_powerapply(apply);
		return -EINVAL;
	}
		
	return 0;
}

static int 
powermode_fail(const int set, const int val)
{
	int r = 0;
	if (cobt_is_monterey()) {
		u8 reg;
		unsigned long flags;

		spin_lock_irqsave(&rtc_lock, flags);
		superio_set_rtc_bank(PC87317_RTC_BANK_APC);
		reg = CMOS_READ(PC87317_APCR6);
		r = reg & 0x80 ? 0 : 1;
		if (set) {
			if (val) {
				reg &= ~0xc0;
			} else {
				reg |= 0x80;
			}
			CMOS_WRITE(reg, PC87317_APCR6);
		}
		superio_set_rtc_bank(PC87317_RTC_BANK_MAIN);
		spin_unlock_irqrestore(&rtc_lock, flags);
	} else if (cobt_is_alpine()) {
		u16 addr;
		u8 reg;
		
		addr = superio_ldev_base(PC87417_DEV_SWC);
		reg = inb(addr + PC87417_SWC_PWONCTL);
		r = reg & 0x20 ? 1 : 0;
		if (set) {
			if (val) {
				reg |= 0x20;
			} else {
				reg &= ~0x20;
			}
			outb(reg, addr + PC87417_SWC_PWONCTL);
		}
	}
	return r;
}

#define NS_GP1_EN0_PME2_E	0x02
#define NS_EVENT_WAKEUP_S345	0x03
static int 
powermode_apply(const int set, const int val)
{
	int r = 0;
	if (cobt_is_monterey()) {
#if 0 /* not supported on monterey hardware */
		u8 reg;
		u16 pmaddr;
		u16 gpeaddr;
	
		/* figure out the PM base address */
		pmaddr = superio_ldev_base(PC87317_DEV_PM);
		/* figure out the GPE base address */
		outb(PC87317_PMDEV_GPELO, pmaddr);
		gpeaddr = inb(pmaddr+1);
		outb(PC87317_PMDEV_GPEHI, pmaddr);
		gpeaddr |= inb(pmaddr+1) << 8;

		reg = inb(gpeaddr + PC87317_GPE_GP1_EN0);
		r = (reg & NS_GP1_EN0_PME2_E) ? 1 : 0;
		if (set) {
			if (val) {
				reg |= NS_GP1_EN0_PME2_E;
			} else {
				reg &= ~NS_GP1_EN0_PME2_E;
			}
			outb(reg, gpeaddr + PC87317_GPE_GP1_EN0);
		}
#endif
		if (set && val) {
			return -1;
		} else {
			return 0;
		}
	} else if (cobt_is_alpine()) {
		u16 pmaddr;
		u8 reg;

		/* power event is attached to GPIOE42 */
		pmaddr = superio_ldev_base(PC87417_DEV_SWC);
		outb(PC87417_SWCWKEVENT_GPIOE42, pmaddr+PC87417_SWC_WKEVENT);
		reg = inb(pmaddr+PC87417_SWC_WKSTATE);
		r = reg&NS_EVENT_WAKEUP_S345 ? 1 : 0;
		if (set) {
			if (val) {
				reg |= NS_EVENT_WAKEUP_S345;
			} else {
				/* disable wakeup */
				reg &= ~NS_EVENT_WAKEUP_S345;
			}
			outb(reg, pmaddr+PC87417_SWC_WKSTATE);
		}
	}
	return r;
}

static void
init_powermode(void)
{
	if (cobt_get_powerapply()) {
		powermode_status = COBALT_POWER_ON;
	} else if (cobt_get_powerfail()) {
		powermode_status = COBALT_POWER_SAME;
	} else {
		powermode_status = COBALT_POWER_OFF;
	}
}

int __init 
cobalt_sysctl_init(void)
{
	init_sysfault();
	init_powermode();

	return 0;
}

#endif /* CONFIG_COBALT */
