/*
 * $Id: systype.c,v 1.23 2001/11/01 00:47:40 thockin Exp $
 * systype.c : routines for figuring out which Cobalt system this is
 *
 * Copyright 2001 Sun Microsystems, Inc.
 *
 * By:  Tim Hockin
 *	Adrian Sun
 *	Duncan Laurie
 *
 * This driver is SMP safe by nature. --TPH
 */

#include <linux/config.h>

#ifdef CONFIG_COBALT

#include <linux/pci.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/module.h>

#include <linux/cobalt.h>
#include <linux/cobalt-systype.h>

/* for easy first-pass analysis */
#if defined(CONFIG_COBALT_GEN_III)
int COBALT_GENERATION_III_DEFINED;
#endif
#if defined(CONFIG_COBALT_GEN_V)
int COBALT_GENERATION_V_DEFINED;
#endif

cobt_sys_t cobt_type = COBT_UNINITIALIZED;
EXPORT_SYMBOL(cobt_type);

#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *proc_systype;
#endif
static int systype_read_proc(char *buf, char **start, off_t pos, int len,
	int *eof, void *x);
static char *systype_str(cobt_sys_t type);

int __init 
cobalt_systype_init(void)
{
	cobalt_systype_probe();

#ifdef CONFIG_PROC_FS
	proc_systype = create_proc_read_entry("systype", 0, 
		proc_cobalt, systype_read_proc, NULL);
	if (!proc_systype) {
		EPRINTK("can't create /proc/cobalt/systype\n");
	}
#endif

	if (cobt_type == COBT_UNKNOWN) {
		printk("Cobalt system type is unknown (trouble will ensue)\n");
		return -1;
	}

	return 0;
}

cobt_sys_t
cobalt_systype_probe(void)
{
	static int init_done = 0;
	struct pci_dev *pdev;

	if (init_done) {
		return cobt_type;
	}
	cobt_type = COBT_UNKNOWN;

#if defined(CONFIG_COBALT_GEN_III)
	/* board identifier for RaQ3/4 vs Qube3 is on the PMU @ 0x7f */
	pdev = pci_find_device(PCI_VENDOR_ID_AL,
		PCI_DEVICE_ID_AL_M7101, NULL);
	if (pdev) {
		/* 
		 * check to see what board we are on
		 * ( RaQ 3, RaQ 4, Qube 3 )
		 */
		unsigned char val;

		/* momentarily set DOGB# to input */
		pci_read_config_byte(pdev, 0x7d, &val);
		pci_write_config_byte(pdev, 0x7d, val & ~0x20);

		/* read the GPIO register */
		pci_read_config_byte(pdev, 0x7f, &val);
		/* RaQ3/4 boards have DOGB (0x20) high, 
		 * Qube3 has DOGB low */ 
		cobt_type = (val & 0x20) ? COBT_PACIFICA : COBT_CARMEL;

		/* change DOGB back to output */
		pci_read_config_byte(pdev, 0x7d, &val);
		pci_write_config_byte(pdev, 0x7d, val | 0x20);

		goto systype_init_done;
	}
#endif
#if defined(CONFIG_COBALT_GEN_V)
	/* is it a gen V ? */
	pdev = pci_find_device(PCI_VENDOR_ID_SERVERWORKS,
		PCI_DEVICE_ID_SERVERWORKS_OSB4, NULL);
	if (pdev) {
		cobt_type = COBT_MONTEREY;
		goto systype_init_done;
	}
	pdev = pci_find_device(PCI_VENDOR_ID_SERVERWORKS,
		PCI_DEVICE_ID_SERVERWORKS_CSB5, NULL);
	if (pdev) {
		cobt_type = COBT_ALPINE;
		goto systype_init_done;
	}
#endif /* CONFIG_COBALT_GEN */

systype_init_done:
	if (cobt_type != COBT_UNKNOWN) {
		init_done = 1;
	}

	/* assign to these, so the compiler shuts up */
#if defined(CONFIG_COBALT_GEN_III)
	COBALT_GENERATION_III_DEFINED = 1;
#endif
#if defined(CONFIG_COBALT_GEN_V)
	COBALT_GENERATION_V_DEFINED = 1;
#endif
	return cobt_type;
}
EXPORT_SYMBOL(cobalt_systype_probe);

#ifdef CONFIG_PROC_FS
static int 
systype_read_proc(char *buf, char **start, off_t pos, int len,
	int *eof, void *x)
{
	int plen = sprintf(buf, "%s\n", systype_str(cobt_type));
	return cobalt_gen_proc_read(buf, plen, start, pos, len, eof);
}
#endif

static char *
systype_str(cobt_sys_t type)
{
	switch (type) {
		case COBT_PACIFICA:
			return "Pacifica";
			break;
		case COBT_CARMEL:
			return "Carmel";
			break;
		case COBT_MONTEREY:
			return "Monterey";
			break;
		case COBT_ALPINE:
			return "Alpine";
			break;
		case COBT_UNKNOWN:
		default:
			return "unknown";
			break;
	}
}
#endif /* CONFIG_COBALT */
