/* $Id: cobalt_rom.c,v 1.3 2001/07/24 03:08:45 thockin Exp $ */
/*
 *
 * This driver implements support for Cobalt Networks systems (x86) flash ROM.
 * These systems have 512k-8MB of flash, mapped in 64k pages at the tippy top
 * of memory.  There is a custom page register that is used to toggle the
 * mapping.  This register changes with different Cobalt hardware.
 *
 * The driver tries to do a detection algorithm to guess what sort of devices
 * are plugged into the sockets.
 *
 * Tim Hockin <thockin@sun.com>
 */

#if 0 //FIXME: TODO:
Cobalt rom:
        erasing - starts to fail after first page
	FIXMEs
	Config.help
	request regions for page registers?
	intel support
		2MB and 1MB sharp, too
	multiple chips
		chip driver for multi-bank chips?
	watch out for multiple readers/writers and read while writing issues..

General:
	fix Config names and descriptions
#endif

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/cobalt-systype.h>
#include <asm/io.h>

#include <linux/mtd/map.h>

#define WINDOW_START 0xffff0000
#define WINDOW_LENGTH 0x10000
#define WINDOW_MASK 0xffff

/* globals, since we only have one flash */
static unsigned long iomap;
static struct pci_dev *pmu;
static int curpage;
static spinlock_t cflash_lock = SPIN_LOCK_UNLOCKED;

/* some Cobalt boards have two banks - for now this is not supported */
static void 
__cflash_page(int page)
{
	int bank = 0;

	if (!pmu) {
		printk("%s: GAHH! PMU is NULL\n", __FUNCTION__);
		return;
	} 
	
	if (curpage == page) {
		return;
	}
	
	if (cobt_is_3k()) {
		__u16 data;
		/* page register is tacked onto the PMU EGPO */
		pci_read_config_word(pmu, 0xb8, &data);
		data &= 0xff80;
		data |= (page & 0x3f) | ((bank&0x1)<<6);
		pci_write_config_word(pmu, 0xb8, data);
	} else if (cobt_is_5k()) {
		/* page register is on OSB4 */
		pci_write_config_byte(pmu, 0x72, 0x08);
		//FIXME: reserve this region?  what magic is 0x900?
		outb(((page & 0x3f) | ((bank&0x1)<<6)), 0x900);
		pci_write_config_byte(pmu, 0x72, 0x0c);
	}
	/* wait to stabilize the page register */
	udelay(2000);
	curpage = page;
	printk("TIM: setting page to %d\n", page);
}


static inline void 
cflash_page(struct map_info *map, unsigned long ofs)
{
	__cflash_page(ofs>>16);
}


static __u8 
cflash_read8(struct map_info *map, unsigned long ofs)
{
	__u8 ret;
	spin_lock(&cflash_lock);
	cflash_page(map, ofs);
	ret = readb(iomap + (ofs & WINDOW_MASK));
	spin_unlock(&cflash_lock);
	return ret;
}

static __u16 
cflash_read16(struct map_info *map, unsigned long ofs)
{
	__u16 ret;
	spin_lock(&cflash_lock);
	cflash_page(map, ofs);
	ret = readw(iomap + (ofs & WINDOW_MASK));
	spin_unlock(&cflash_lock);
	return ret;
}

static __u32 
cflash_read32(struct map_info *map, unsigned long ofs)
{
	__u32 ret;
	spin_lock(&cflash_lock);
	cflash_page(map, ofs);
	ret = readl(iomap + (ofs & WINDOW_MASK));
	spin_unlock(&cflash_lock);
	return ret;
}

static void 
cflash_cp_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	while(len) {
		unsigned long thislen = len;
		if (len > (WINDOW_LENGTH - (from & WINDOW_MASK)))
			thislen = WINDOW_LENGTH-(from & WINDOW_MASK);
		
		spin_lock(&cflash_lock);
		cflash_page(map, from);
		memcpy_fromio(to, iomap + (from & WINDOW_MASK), thislen);
		spin_unlock(&cflash_lock);
		to += thislen;
		from += thislen;
		len -= thislen;
	}
}

static void 
cflash_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	spin_lock(&cflash_lock);
	cflash_page(map, adr);
	writeb(d, iomap + (adr & WINDOW_MASK));
	spin_unlock(&cflash_lock);
}

static void 
cflash_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	spin_lock(&cflash_lock);
	cflash_page(map, adr);
	writew(d, iomap + (adr & WINDOW_MASK));
	spin_unlock(&cflash_lock);
}

static void 
cflash_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	spin_lock(&cflash_lock);
	cflash_page(map, adr);
	writel(d, iomap + (adr & WINDOW_MASK));
	spin_unlock(&cflash_lock);
}

static void 
cflash_cp_to(struct map_info *map, unsigned long to, const void *from, 
	ssize_t len)
{
	while(len) {
		unsigned long thislen = len;
		if (len > (WINDOW_LENGTH - (to & WINDOW_MASK)))
			thislen = WINDOW_LENGTH-(to & WINDOW_MASK);
		
		spin_lock(&cflash_lock);
		cflash_page(map, to);
		memcpy_toio(iomap + (to & WINDOW_MASK), from, thislen);
		spin_unlock(&cflash_lock);
		to += thislen;
		from += thislen;
		len -= thislen;
	}
}

static struct map_info cflash_map = {
	name: "Cobalt Networks Flash ROM",
	size: 0,
	buswidth: 1,
	read8: cflash_read8,
	read16: cflash_read16,
	read32: cflash_read32,
	copy_from: cflash_cp_from,
	write8: cflash_write8,
	write16: cflash_write16,
	write32: cflash_write32,
	copy_to: cflash_cp_to,
};

static struct mtd_info *cflash_mtd;

#if LINUX_VERSION_CODE < 0x20212 && defined(MODULE)
#define init_cflash init_module
#define cleanup_cflash cleanup_module
#endif

void 
cleanup_cflash(void)
{
	/* reset page to 0 */
	__cflash_page(0);
	
	if (cflash_mtd) {
		del_mtd_device(cflash_mtd);
		map_destroy(cflash_mtd);
	}

	iounmap((void *)iomap);
}

int __init 
init_cflash(void)
{
	/* figure out how to do paging, and allocate anything we need */
	if (cobt_is_3k()) {
		pmu = pci_find_device(PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_M7101, 
			NULL);
	} else if (cobt_is_5k()) {
		pmu = pci_find_device(PCI_VENDOR_ID_SERVERWORKS,
			PCI_DEVICE_ID_SERVERWORKS_OSB4, NULL);
	}
	if (!pmu) {
		return -ENODEV;
	}

	/* ROM pages live in the high 64k of memory */
	iomap = (unsigned long)ioremap(WINDOW_START, WINDOW_LENGTH);
	if (!iomap) {
		printk("Failed to ioremap memory region\n");
		return -EIO;
	}
	
	/* set page to 0 */
	__cflash_page(0);

	/* probe for chips - we know what kinds */
	cflash_mtd = do_map_probe("jedec_probe", &cflash_map);
	if (cflash_mtd) {
		cflash_mtd->module = THIS_MODULE;
		cflash_map.size = cflash_mtd->size;
		add_mtd_device(cflash_mtd);
	} else {
		cleanup_cflash();
		return -ENXIO;
	}	  

#if 0 /?FIXME:
	// Do an autoprobe sequence
	if (check_region(PAGE_IO,1) != 0) {
		printk("5066: Page Register in Use\n");
		return -EAGAIN;
	}
	request_region(PAGE_IO,1,"Octagon SSD");
#endif
	
	return 0;
}

module_init(init_cflash);
module_exit(cleanup_cflash);
